#include <stdio.h>

unsigned char * strings[] = {
/* 00 */	(unsigned char *)"3[APA3A] tiny proxy v0.4.5b stringtable file",
/* 01 */	(unsigned char *)"3proxy",
/* 02 */	(unsigned char *)"3[APA3A]tiny proxy",
/* 03 */	(unsigned char *)"0.4.5b",
/* 04 */	NULL,
/* 05 */	(unsigned char *)"SERVR",
/* 06 */	(unsigned char *)"PROXY",
/* 07 */	(unsigned char *)"TCPPM",
/* 08 */	(unsigned char *)"POP3P",
/* 09 */	(unsigned char *)"SOCK4",
/* 10 */	(unsigned char *)"SOCK5",
/* 11 */	(unsigned char *)"UDPPM",
/* 12 */	(unsigned char *)"SOCKS",
/* 13 */	(unsigned char *)"SOC45",
/* 14 */	(unsigned char *)"ADMIN",
/* 15 */	(unsigned char *)"/usr/local/etc/3proxy.cfg",
/* 16 */	(unsigned char *)"3proxy.cfg",
/* 17 */	NULL,
/* 18 */	NULL,
/* 19 */	NULL,
/* 20 */	NULL,
/* 21 */	NULL,
/* 22 */	NULL,
/* 23 */	NULL,
/* 24 */	NULL,
/* 25 */	NULL,
/* 26 */	NULL,
/* 27 */	NULL,
/* 28 */	NULL,
/* 29 */	NULL,
/* 30 */	NULL,
/* 31 */	NULL,
/* 32 */	NULL,
/* 33 */	NULL,
/* 34 */	NULL,
/* 35 */	NULL,
/* 36 */	NULL,
/* 37 */	NULL,
/* 38 */	NULL,
/* 39 */	NULL,
};

int constants[] = {0,0};
